"""The Django ModelFactories for Amsterdam Schema mock data.

When ModelFactories are generated with
:func:`~schematools.contrib.django.factories.model_mocker_factory`,
they all inherit from
:class:`~schematools.contrib.django.mockers.DynamicModelFactory`
to have a common interface.

Note that the Models and ModelFactories are bootstrapped from
`~schematools.contrib.django.factories.py`, so to circumvent this naming clash,
we place the ModelFactory logic in `~schematools.contrib.django.mockers.py`
as the primary aim of the ModelFactory is to generate Mock data for development
and testing purposes.
"""

from factory.django import DjangoModelFactory


class DynamicModelMocker(DjangoModelFactory):
    """Base class to tag and detect dynamically generated model mockers."""

    @classmethod
    def _setup_next_sequence(cls):
        # Instead of defaulting to starting at 0, we start at > 0
        # this is more appropriate for database keys
        # NB This classmethod can be replaced in the dango mgm command
        # to be able to start at a higher sequence number.
        return 1



DATASET_TMPL = {
    "type": "dataset",
    "id": None,
    "title": None,
    "status": "beschikbaar",
    "description": None,
    "crs": "EPSG:28992",
    "auth": "OPENBAAR",
    "authorizationGrantor": "n.v.t.",
    "owner": "Gemeente Amsterdam",
    "creator": "bronhouder onbekend",
    "publisher": "Datateam xyz",
    "tables": [],
}

# The display field will be hard-coded as 'id', because we cannot know this value
# by purely inspecting the postgresql db.
TABLE_TMPL = {
    "id": None,
    "type": "table",
    "version": "0.0.1",
    "schema": {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "type": "object",
        "additionalProperties": False,
        "required": [],
        "display": "id",
        "properties": {
            "schema": {
                "$ref": "https://schemas.data.amsterdam.nl/schema@v1.3.0#/definitions/schema"
            },
        },
    },
}



import json
import re

import pytest

from schematools import ckan
from schematools.types import DatasetSchema


@pytest.mark.parametrize(
    "name",
    [
        "afval",
        "afvalwegingen",
        "brk",
        "brp",
        "composite_key",
        "ggwgebieden",
        "hr",
        "meetbouten",
        "meldingen",
        "parkeervakken",
        "stadsdelen",
    ],
)
def test_convert(here, name):
    filename = here / f"files/datasets/{name}.json"
    with open(filename) as f:
        schema = DatasetSchema.from_dict(json.load(f))

    data = ckan.from_dataset(schema, name)

    for key in ["identifier", "title"]:
        assert data.get(key) and re.match(r"^[a-z0-9_-]", data[key])
    for key in ["language", "theme"]:
        assert isinstance(data.get(key), list) and data[key]
    assert data["identifier"]
    assert data["identifier"] == data["url"]
    assert data["notes"]



from django.db.models import QuerySet
from django.db.models.query import ModelIterable


class DatasetIterable(ModelIterable):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        from .loaders import DatabaseSchemaLoader

        self._dataset_collection = DatabaseSchemaLoader().dataset_collection

    def __iter__(self):
        """Inject a shared dataset collection with the results of this queryset."""
        for obj in super().__iter__():
            obj._dataset_collection = self._dataset_collection
            yield obj


class DatasetQuerySet(QuerySet):
    """Extra ORM methods for the Dataset model.

    All datasets that were retrieved with the same queryset are linked together,
    so they can resolve relations between each other.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._iterable_class = DatasetIterable

    def db_enabled(self):
        """Return all datasets for which models should be created."""
        return self.filter(enable_db=True)

    def endpoint_enabled(self):
        """Return the datasets where data is retrieved from remote API's."""
        return self.filter(enable_db=False, endpoint_url__isnull=False)

    def api_enabled(self):
        """Return only datasets that should get an API."""
        return self.filter(enable_api=True)



from __future__ import annotations

import logging
import warnings
from pathlib import Path

from more_ds.network.url import URL

from schematools import types
from schematools.loaders import FileSystemSchemaLoader, URLSchemaLoader

logger = logging.getLogger(__name__)
_shared_fs_loader = None

warnings.warn(
    "Using schematools.utils is deprecated, use the `schematools.loaders` module instead.",
    DeprecationWarning,
)


def dataset_schemas_from_url(
    schemas_url: URL | str,
    dataset_name: str | None = None,
    prefetch_related: bool = False,
) -> dict[str, types.DatasetSchema]:
    warnings.warn(
        "Using dataset_schema_from_url() is deprecated, "
        "use `schematools.loaders.URLSchemaLoader` or `get_schema_loader()` instead.",
        DeprecationWarning,
    )

    loader = URLSchemaLoader(URL(schemas_url))
    if dataset_name:
        return {
            loader.get_dataset_path(dataset_name): loader.get_dataset(
                dataset_name, prefetch_related=prefetch_related
            )
        }
    else:
        return loader.get_all_datasets()


def dataset_schema_from_url(
    schemas_url: URL | str,
    dataset_name: str,
    prefetch_related: bool = False,
) -> types.DatasetSchema:
    """Fetch a dataset schema from a remote file."""
    warnings.warn(
        "Using dataset_schema_from_url() is deprecated, "
        "use `schematools.loaders.URLSchemaLoader` or `get_schema_loader()` instead.",
        DeprecationWarning,
    )

    loader = URLSchemaLoader(schemas_url)
    return loader.get_dataset(dataset_name, prefetch_related=prefetch_related)


def schemas_from_url(base_url: URL | str, data_type: type) -> dict[str, types.DatasetSchema]:
    """Fetch all schema definitions from a remote file.

    The URL could be ``https://schemas.data.amsterdam.nl/datasets/``
    """
    warnings.warn(
        "Using schemas_from_url() is deprecated, "
        "use `schematools.loaders.URLSchemaLoader().get_all_datasets()` instead.",
        DeprecationWarning,
    )
    if data_type is not types.DatasetSchema:
        raise TypeError("schemas_from_url() only worked with DatasetSchema")

    loader = URLSchemaLoader(base_url)
    return loader.get_all_datasets()


def schema_from_url(
    base_url: URL | str,
    data_type: type,
    dataset_id: str,
    prefetch_related: bool = False,
) -> types.DatasetSchema:
    """Fetch schema definitions from a remote file for a single dataset.

    The URL could be ``https://schemas.data.amsterdam.nl/datasets/``
    """
    warnings.warn(
        "Using schema_from_url() is deprecated, "
        "use `schematools.loaders.URLSchemaLoader().get_dataset()` instead.",
        DeprecationWarning,
    )
    if data_type is not types.DatasetSchema:
        raise TypeError("schema_from_url() only worked with DatasetSchema")

    loader = URLSchemaLoader(base_url)
    return loader.get_dataset(dataset_id, prefetch_related=prefetch_related)


def dataset_schema_from_path(dataset_path: Path | str) -> types.DatasetSchema:
    """Read a dataset schema from the filesystem.

    Args:
        dataset_path: Filesystem path to the dataset.
    """
    warnings.warn(
        "Using dataset_schema_from_path() is deprecated, "
        "use `schematools.loaders.FileSystemSchemaLoader` instead.",
        DeprecationWarning,
    )

    # The filesystem loader is stored in a global instance,
    # so all loaded datasets are part of the same "dataset collection".
    global _shared_fs_loader
    if _shared_fs_loader is None:
        try:
            root = FileSystemSchemaLoader.get_root(dataset_path)
        except ValueError:
            root = dataset_path.parent
        _shared_fs_loader = FileSystemSchemaLoader(root)

    return _shared_fs_loader.get_dataset_from_file(dataset_path)


def dataset_schemas_from_schemas_path(root: Path | str) -> dict[str, types.DatasetSchema]:
    """Read all dataset schemas from a directory tree.

    Args:
        root: Path of a directory containing dataset schemas in subdirs.
    """
    warnings.warn(
        "Using dataset_schema_from_path() is deprecated, "
        "use `schematools.loaders.FileSystemSchemaLoader.get_all_datasets()` instead.",
        DeprecationWarning,
    )

    loader = FileSystemSchemaLoader(root)
    return loader.get_all_datasets()


def publishers_from_url(base_url: URL | str) -> dict[str, list[types.Publisher]]:
    """
    The URL could be ``https://schemas.data.amsterdam.nl/publishers/``
    """

    loader = URLSchemaLoader(base_url)
    return loader.get_all_publishers()



"""Module implementing an event processor, that processes full events."""
from __future__ import annotations

import logging
import re
from collections import defaultdict
from dataclasses import dataclass

import orjson
from sqlalchemy import Table, inspect
from sqlalchemy.engine import Connection

from schematools.events import metadata
from schematools.exceptions import DatasetTableNotFound
from schematools.factories import tables_factory
from schematools.importer.base import BaseImporter
from schematools.loaders import get_schema_loader
from schematools.naming import to_snake_case
from schematools.types import DatasetFieldSchema, DatasetSchema, DatasetTableSchema

# Enable the sqlalchemy logger by uncommenting the following 2 lines to debug SQL related issues
# logging.basicConfig()
# logging.getLogger("sqlalchemy.engine").setLevel(logging.INFO)

logger = logging.getLogger(__name__)

# Configuration information to map the event-type
# to the following fields:
# - db_operation_name (INSERT/UPDATE)
# - needs_select: generated SQL needs to select a record

EVENT_TYPE_MAPPINGS = {
    "ADD": ("insert", False),
    "MODIFY": ("update", True),
    "DELETE": ("delete", True),
}

FULL_LOAD_TABLE_POSTFIX = "_full_load"


@dataclass
class UpdateParentTableConfiguration:
    parent_schema_table: DatasetTableSchema
    parent_table: Table
    relation_name: str

    @property
    def parent_id_field(self):
        return (
            self.parent_table.c.id
            if self.parent_schema_table.has_composite_key
            else getattr(self.parent_table.c, self.parent_schema_table.identifier[0])
        )

    def parent_id_value(self, prepared_row: dict) -> str:
        return ".".join(
            [
                str(prepared_row[to_snake_case(f"{self.parent_schema_table.id}_{fn}")])
                for fn in self.parent_schema_table.identifier
            ]
        )

    @property
    def parent_fields(self) -> list[str]:
        pattern = re.compile(f"{self.relation_name}_[a-zA-Z]+")
        return [f.name for f in self.parent_table.columns if re.match(pattern, f.name) is not None]

    def parent_field_values(self, data: dict) -> dict:
        return {k: data[k] for k in self.parent_fields}


@dataclass
class RunConfiguration:
    check_existence_on_add: bool = False
    process_events: bool = True
    execute_after_process: bool = True

    """Whether or not to update the table. This should usually be the case, but this variable can
    be set to False when the table does not exist. This can happen for certain relation events
    for which no relation table is defined in the database. In this case we still want to update
    the relation columns in the parent table.

    """
    update_table: bool = True

    table: Table = None
    table_name: str = None
    schema_table: DatasetTableSchema = None

    update_parent_table_configuration: UpdateParentTableConfiguration = None

    nested_table_fields: list[DatasetFieldSchema] = None


class LastEventIds:
    BENK_DATASET = "benk"
    LASTEVENTIDS_TABLE = "lasteventids"

    def __init__(self):
        loader = get_schema_loader()
        self.dataset = loader.get_dataset(self.BENK_DATASET)
        self.lasteventids_table = self.dataset.get_table_by_id(self.LASTEVENTIDS_TABLE)
        self.lasteventid_column = self.lasteventids_table.get_field_by_id("lastEventId")

        self.cache = defaultdict(int)

    def is_event_processed(self, conn, table_name: str, event_id: int) -> bool:
        last_eventid = self.get_last_eventid(conn, table_name)

        return False if last_eventid is None else last_eventid >= event_id

    def update_eventid(self, conn, table_name: str, event_id: int | None):
        value = event_id if event_id is not None else "NULL"
        conn.execute(
            f"INSERT INTO {self.lasteventids_table.db_name} "  # noqa: S608  # nosec: B608
            f"VALUES ('{table_name}', {value}) "
            f'ON CONFLICT ("table") DO UPDATE SET {self.lasteventid_column.db_name} = {value}'
        )
        self.cache[table_name] = event_id

    def get_last_eventid(self, conn, table_name: str) -> int | None:
        if table_name in self.cache:
            return self.cache[table_name]

        res = conn.execute(
            f"SELECT {self.lasteventid_column.db_name} "  # noqa: S608
            f"FROM {self.lasteventids_table.db_name} "  # noqa: S608
            f"WHERE \"table\" = '{table_name}'"  # noqa: S608
        ).fetchone()
        last_eventid = None if res is None else res[0]
        self.cache[table_name] = last_eventid
        return last_eventid

    def copy_lasteventid(self, conn, from_table_name: str, to_table_name: str):
        eventid = self.get_last_eventid(conn, from_table_name)
        self.update_eventid(conn, to_table_name, eventid)

    def clear_cache(self):
        self.cache = defaultdict(int)


class EventsProcessor:
    """The core event processing class.

    It needs to be initialised once
    with configuration (datasets) and a db connection.
    Once initialised, the process_event() method is able to
    process incoming events.
    The database actions are done using SQLAlchemy Core. So,
    a helper function `tables_factory()` is used to created the
    SA Tables that are needed during the processing of the events.
    """

    def __init__(
        self,
        datasets: list[DatasetSchema],
        connection: Connection,
        local_metadata=None,
        truncate=False,
    ) -> None:
        """Construct the event processor.

        Args:
            datasets: list of DatasetSchema instances. Usually dataset tables
                have relations to tables.
                If these target tables are in different datasets,
                these dataset also need to be provided.
            srid: coordinate system
            local_metadata: SQLAlchemy metadata object, only needs to be provided
                in unit tests.
            truncate: indicates if the relational tables need to be truncated
        """
        self.lasteventids = LastEventIds()
        benk_dataset = self.lasteventids.lasteventids_table.dataset
        self.datasets = {benk_dataset.id: benk_dataset} | {ds.id: ds for ds in datasets}
        self.conn = connection
        _metadata = local_metadata or metadata  # mainly for testing
        _metadata.bind = connection.engine
        inspector = inspect(connection.engine)
        self.tables = {}
        self.full_load_tables = defaultdict(dict)
        for dataset_id, dataset in self.datasets.items():
            base_tables_ids = {dataset_table.id for dataset_table in dataset.tables}
            self.tables[dataset_id] = tfac = {
                # As quick workaround to map table identifiers with the existing event streams,
                # the code of this file works with snake cased identifiers.
                # The remaining code (e.g. BaseImporter) already works directly with table.id.
                to_snake_case(table_id): table
                for table_id, table in tables_factory(dataset, metadata=_metadata).items()
            }
            self.geo_fields = defaultdict(lambda: defaultdict(list))
            for table_id, table in tfac.items():
                if not inspector.has_table(table.name):
                    table.create()
                elif truncate:
                    with self.conn.begin():
                        self.conn.execute(table.delete())
                # self.has_composite_key = dataset_table.has_composite_key
                # skip the generated nm tables
                if table_id not in base_tables_ids:
                    continue
                for field in dataset.get_table_by_id(table_id).fields:
                    if field.is_geo:
                        self.geo_fields[dataset_id][table_id].append(field)

    def _flatten_event_data(self, event_data: dict) -> dict:
        result = {}
        for k, v in event_data.items():
            if isinstance(v, dict):
                flattened = self._flatten_event_data(v)
                for k2, v2 in flattened.items():
                    result[f"{k}_{k2}"] = v2
            else:
                result[k] = v
        return result

    def _get_full_load_tables(
        self, dataset_id: str, table_id: str
    ) -> tuple[Table, DatasetTableSchema]:
        try:
            # Cached
            table, schema_table = self.full_load_tables[dataset_id][table_id]
        except KeyError:
            # Initialise fresh tables for full load.
            dataset_schema: DatasetSchema = self.datasets[dataset_id]
            tables = dataset_schema.get_tables(include_nested=True, include_through=True)

            importer = BaseImporter(dataset_schema, self.conn.engine, logger)
            for schema_table in tables:
                this_table_id = schema_table.id
                db_table_name = schema_table.db_name_variant(postfix=FULL_LOAD_TABLE_POSTFIX)

                importer.generate_db_objects(
                    this_table_id,
                    db_schema_name=dataset_id,
                    db_table_name=db_table_name,
                    is_versioned_dataset=importer.is_versioned_dataset,
                    ind_extra_index=False,
                    ind_create_pk_lookup=False,
                )

                table = importer.tables[this_table_id]
                self.full_load_tables[dataset_id][to_snake_case(this_table_id)] = (
                    table,
                    schema_table,
                )
            try:
                table, schema_table = self.full_load_tables[dataset_id][table_id]
            except KeyError:
                # Happens when there is no relation table to update for a given relation, for ex.
                raise DatasetTableNotFound()

        return table, schema_table

    def _before_process(self, run_configuration: RunConfiguration, event_meta: dict) -> None:
        if not run_configuration.update_table:
            return
        if event_meta.get("full_load_sequence", False):

            if event_meta.get("first_of_sequence", False):
                self.conn.execute(f"TRUNCATE {run_configuration.table.fullname}")

    def _after_process(self, run_configuration: RunConfiguration, event_meta: dict):
        if not run_configuration.update_table:
            return

        if event_meta.get("full_load_sequence", False) and event_meta.get(
            "last_of_sequence", False
        ):
            dataset_id = event_meta["dataset_id"]
            table_id = event_meta["table_id"]

            table_ids_to_replace = [
                table_id,
            ] + [to_snake_case(f.nested_table.id) for f in run_configuration.nested_table_fields]

            logger.info("End of full load sequence. Replacing active table.")
            with self.conn.begin():

                full_load_tables = []
                for t_id in table_ids_to_replace:
                    table_to_replace = self.tables[dataset_id][to_snake_case(t_id)]
                    full_load_table, full_load_schema_table = self._get_full_load_tables(
                        dataset_id, to_snake_case(t_id)
                    )
                    full_load_tables.append(full_load_table)

                    fieldnames = ", ".join(
                        [field.db_name for field in full_load_schema_table.get_db_fields()]
                    )

                    self.conn.execute(f"TRUNCATE {table_to_replace.fullname}")
                    self.conn.execute(
                        f"INSERT INTO {table_to_replace.fullname} ({fieldnames}) "  # noqa: S608
                        f"SELECT {fieldnames} FROM {full_load_table.fullname}"  # noqa: S608
                    )
                if run_configuration.update_parent_table_configuration:
                    self._update_parent_table_bulk(run_configuration)

                for full_load_table in full_load_tables:
                    self.conn.execute(f"DROP TABLE {full_load_table.fullname} CASCADE")

                # Copy full_load lasteventid to active table and set full_load lasteventid to None
                self.lasteventids.copy_lasteventid(
                    self.conn, run_configuration.table_name, table_to_replace.name
                )
                self.lasteventids.update_eventid(self.conn, run_configuration.table_name, None)

            self.full_load_tables[dataset_id].pop(to_snake_case(table_id))

    def _prepare_row(
        self,
        run_configuration: RunConfiguration,
        event_meta: dict,
        event_data: dict,
        schema_table: DatasetTableSchema,
    ) -> dict:
        dataset_id = event_meta["dataset_id"]
        table_id = event_meta["table_id"]

        row = self._flatten_event_data(event_data)

        # Set null values for missing fields after flattening (e.g. when a 1-1 relation is empty)
        # Only applies to events for which we have a table
        if schema_table:
            row |= {
                f.db_name: None
                for f in schema_table.get_fields(include_subfields=True)
                if f.db_name not in row and f.db_name not in ("id", "schema")
            }

        for geo_field in self.geo_fields[dataset_id][table_id]:
            row_key = to_snake_case(geo_field.name)
            geo_value = row.get(row_key)
            if geo_value is not None and not geo_value.startswith("SRID"):
                row[row_key] = f"SRID={geo_field.srid};{geo_value}"

        if run_configuration.update_table:
            identifier = schema_table.identifier
            id_value = ".".join(str(row[to_snake_case(fn)]) for fn in identifier)
            row["id"] = id_value
        return row

    def _update_parent_table(
        self,
        configuration: UpdateParentTableConfiguration,
        event_meta: dict,
        prepared_row: dict,
    ):
        # Have 1:n relation. We need to update the relation columns in the parent table as
        # well. Skips this for n:m relations (schematable.parent_table_field.relation only
        # returns 1:n relations)
        stmt = configuration.parent_table.update().where(
            configuration.parent_id_field == configuration.parent_id_value(prepared_row)
        )

        update_row = (
            configuration.parent_field_values(prepared_row)
            if event_meta["event_type"] != "DELETE"
            else {k: None for k in configuration.parent_fields}
        )

        self.conn.execute(stmt, update_row)

    def _update_parent_table_bulk(self, run_configuration: RunConfiguration):
        update_parent_table_config = run_configuration.update_parent_table_configuration
        parent_table_ref_id = f"{run_configuration.schema_table.id.split('_')[0]}_id"

        if len(update_parent_table_config.parent_fields) > 1:
            # Adds parentheses
            set_fields = (
                f'({", ".join(update_parent_table_config.parent_fields)}) = '
                f'(s.{", s.".join(update_parent_table_config.parent_fields)})'
            )
        else:
            # No parentheses, because only one field and Postgres will fail on this
            set_fields = (
                f"{update_parent_table_config.parent_fields[0]} = "
                f"s.{update_parent_table_config.parent_fields[0]}"
            )

        query = f"""
        UPDATE {update_parent_table_config.parent_table.fullname} p
        SET {set_fields}
        FROM {run_configuration.table.fullname} s
        WHERE p.{update_parent_table_config.parent_id_field.name} = s.{parent_table_ref_id}
        ;
        """  # noqa: S608

        self.conn.execute(query)

    def _process_row(
        self, run_configuration: RunConfiguration, event_meta: dict, event_data: dict
    ) -> None:
        """Process one row of data.

        Args:
            run_configuration: Configuration for the current run
            event_meta: Metadata about the event
            event_data: Data containing the fields of the event
        """
        table = run_configuration.table
        schema_table = run_configuration.schema_table

        if self.lasteventids.is_event_processed(
            self.conn, run_configuration.table_name, event_meta["event_id"]
        ):
            logger.warning("Event with id %s already processed. Skipping.", event_meta["event_id"])
            return

        row = self._prepare_row(run_configuration, event_meta, event_data, schema_table)
        id_value = row["id"]

        event_type = event_meta["event_type"]

        if (
            run_configuration.check_existence_on_add
            and event_type == "ADD"
            and self._row_exists_in_database(run_configuration, id_value)
        ):
            logger.info("Row with id %s already exists in database. Skipping.", row["id"])
            return

        db_operation = None
        if run_configuration.update_table:
            db_operation_name, needs_select = EVENT_TYPE_MAPPINGS[event_type]
            db_operation = getattr(table, db_operation_name)()

            if needs_select:
                id_field = (
                    table.c.id
                    if schema_table.has_composite_key
                    else getattr(table.c, schema_table.identifier[0])
                )
                db_operation = db_operation.where(id_field == id_value)
        with self.conn.begin():
            if run_configuration.update_table:
                self.conn.execute(db_operation, row)
                self._update_nested_tables(run_configuration, row, event_type, id_value)
            self.lasteventids.update_eventid(
                self.conn, run_configuration.table_name, event_meta["event_id"]
            )

            if run_configuration.update_parent_table_configuration:
                self._update_parent_table(
                    run_configuration.update_parent_table_configuration,
                    event_meta,
                    row,
                )

    def _update_nested_tables(
        self, run_configuration: RunConfiguration, prepared_row: dict, event_type: str, id_value
    ):
        is_delete = event_type == "DELETE"

        for field in run_configuration.nested_table_fields:
            schema_table: DatasetTableSchema = field.nested_table
            table = self.tables[run_configuration.schema_table.dataset.id][
                to_snake_case(schema_table.id)
            ]

            self.conn.execute(table.delete().where(table.c.parent_id == id_value))

            if is_delete:
                continue

            if value := prepared_row.get(field.id, []):
                if rows := self._prepare_nested_rows(field, value, id_value):
                    self.conn.execute(table.insert(), rows)

    def _prepare_nested_rows(self, field: DatasetFieldSchema, value: list, parent_id_value: str):
        return [
            {
                "parent_id": parent_id_value,
            }
            | {subfield.db_name: v[subfield.db_name] for subfield in field.subfields}
            for v in value
        ]

    def _update_nested_tables_bulk(self, run_configuration: RunConfiguration, rows: list[dict]):

        for field in run_configuration.nested_table_fields:
            schema_table: DatasetTableSchema = field.nested_table
            table, _ = self._get_full_load_tables(
                run_configuration.schema_table.dataset.id, to_snake_case(schema_table.id)
            )

            nested_rows = []
            for row in rows:
                nested_rows += self._prepare_nested_rows(
                    field, row.get(to_snake_case(field.id)) or [], row["id"]
                )
            if nested_rows:
                self.conn.execute(table.insert(), nested_rows)

    def _row_exists_in_database(self, run_configuration: RunConfiguration, id_value: str):
        table = run_configuration.table
        schema_table = run_configuration.schema_table
        id_field = (
            table.c.id
            if schema_table.has_composite_key
            else getattr(table.c, schema_table.identifier[0])
        )

        with self.conn.begin():
            res = self.conn.execute(table.select().where(id_field == id_value))
            try:
                next(res)
            except StopIteration:
                return False
            return True

    def _table_empty(self, table: Table):
        with self.conn.begin():
            res = self.conn.execute(table.select())
            try:
                next(res)
            except StopIteration:
                return True
            return False

    def _get_run_configuration(
        self, first_event_meta: dict, last_event_meta: dict, recovery_mode: bool
    ) -> RunConfiguration:
        run_configuration = RunConfiguration()
        dataset_id = first_event_meta["dataset_id"]
        table_id = first_event_meta["table_id"]

        try:
            if first_event_meta.get("full_load_sequence", False):
                table, schema_table = self._get_full_load_tables(
                    dataset_id, to_snake_case(table_id)
                )
            else:
                schema_table = self.datasets[dataset_id].get_table_by_id(table_id)
                table = self.tables[dataset_id][to_snake_case(table_id)]
            run_configuration.table = table
            run_configuration.schema_table = schema_table
            run_configuration.table_name = table.name
            run_configuration.nested_table_fields = [
                f for f in schema_table.fields if f.is_nested_table
            ]

            if schema_table.has_parent_table and schema_table.parent_table_field.relation:
                run_configuration.update_parent_table_configuration = (
                    UpdateParentTableConfiguration(
                        parent_schema_table=schema_table.parent_table,
                        parent_table=self.tables[dataset_id][schema_table.parent_table.id],
                        relation_name=to_snake_case(schema_table.parent_table_field.shortname),
                    )
                )
        except DatasetTableNotFound as exc:
            # Check if relation table, if so continue
            parent_table_id, *field_id = table_id.split("_")
            field_id = "_".join(field_id)

            parent_table = self.datasets[dataset_id].get_table_by_id(parent_table_id)
            field = parent_table.get_field_by_id(field_id)
            if field.get("relation"):
                logger.info(
                    "Relation %s.%s has no table. Will only update parent table.",
                    dataset_id,
                    table_id,
                )
                run_configuration.update_table = False
                run_configuration.update_parent_table_configuration = (
                    UpdateParentTableConfiguration(
                        parent_schema_table=parent_table,
                        parent_table=self.tables[dataset_id][parent_table_id],
                        relation_name=to_snake_case(field.shortname),
                    )
                )
                run_configuration.table_name = to_snake_case(table_id)

            else:
                raise exc

        if recovery_mode:
            self._recover(run_configuration, first_event_meta, last_event_meta)
        return run_configuration

    def _recover(
        self, run_configuration: RunConfiguration, first_event_meta: dict, last_event_meta: dict
    ):
        # If a message is redelivered, we need to enter recovery mode. Redelivery means something
        # has gone wrong
        # somewhere. We need to get back to a consistent state.
        # The actions to take in recovery mode depend on the type of message:
        # 1. full_load_sequence = False: This was a regular update event. Can be of any type
        #     (ADD/MODIFY/DELETE). MODIFY
        #     and DELETE are idempotent, but ADD is not. We need to check if the data is already
        #     in the database before
        #     trying to add it again.
        # 2. full_load_sequence = True with first_of_sequence = True: This should not be a problem.
        #     The first_of_sequence
        #     causes the table to be truncated, so we can just continue as normal. No need to check
        #     for existence.
        # 3. full_load_sequence = True with first_of_sequence = False and last_of_sequence = False:
        #     We should check
        #     for existence before adding event data to the table. Because first_of_sequence and
        #     last_of_sequence are
        #     both False, there are no other possible side effects to consider.
        # 4. full_load_sequence = True with first_of_sequence = False and last_of_sequence = True:
        #    If the target table
        #    is empty, we know that after_process was executed and that this message was handled
        #    correctly the first
        #    time it got delivered (because first_of_sequence = False, there should already have
        #    been data in the
        #    table). In that case we can ignore everything in this message; skip processing the
        #    events and skip the
        #    after_process step (4a). If the target table is not empty, we know that after_process
        #    was not executed. We
        #    should process the events and check for existence of the first event. After that we
        #    should execute
        #    after_process (4b).
        if first_event_meta.get("full_load_sequence", False):
            if first_event_meta.get("first_of_sequence", False):
                # Case 2.
                pass
            elif not last_event_meta.get("last_of_sequence", False):
                # Case 3.
                run_configuration.check_existence_on_add = True
            else:
                # Case 4.
                if self._table_empty(run_configuration.table):
                    # Case 4a.
                    run_configuration.execute_after_process = False
                    run_configuration.process_events = False
                else:
                    # Case 4b.
                    run_configuration.check_existence_on_add = True
        else:
            # Case 1.
            run_configuration.check_existence_on_add = True

    def process_event(self, event_meta: dict, event_data: dict, recovery_mode: bool = False):
        self.process_events([(event_meta, event_data)], recovery_mode)

    def process_events(self, events: list[tuple[dict, dict]], recovery_mode: bool = False):
        if len(events) == 0:
            return

        first_event_meta, last_event_meta = events[0][0], events[-1][0]
        run_configuration = self._get_run_configuration(
            first_event_meta, last_event_meta, recovery_mode
        )

        self._before_process(run_configuration, first_event_meta)

        if run_configuration.process_events:
            if first_event_meta.get("full_load_sequence", False):
                # full_load_sequence only contains add events. Take more efficient shortcut.
                self._process_bulk_adds(run_configuration, events)
            else:
                for event_meta, event_data in events:
                    self._process_row(run_configuration, event_meta, event_data)

        if run_configuration.execute_after_process:
            self._after_process(run_configuration, last_event_meta)

    def _process_bulk_adds(
        self, run_configuration: RunConfiguration, events: list[tuple[dict, dict]]
    ):
        first = True
        rows = []
        last_eventid = None

        for event_meta, event_data in events:
            if event_meta["event_type"] != "ADD":
                raise Exception("This method should only be called when processing ADD events.")

            row = self._prepare_row(
                run_configuration, event_meta, event_data, run_configuration.schema_table
            )
            if (
                run_configuration.check_existence_on_add
                and first
                and run_configuration.update_table
            ):
                if self._row_exists_in_database(run_configuration, row["id"]):
                    logger.info("Skip bulk adds, as the first row already exists in the database.")
                    return
                first = False

            # Note: This has some overlap with the recovery mode
            if self.lasteventids.is_event_processed(
                self.conn, run_configuration.table_name, event_meta["event_id"]
            ):
                logger.warning(
                    "Skip event with id %s, as the event has already been processed.",
                    event_meta["event_id"],
                )
                continue

            rows.append(row)
            last_eventid = event_meta["event_id"]

        if len(rows) == 0:
            return

        with self.conn.begin():
            if run_configuration.update_table:
                self.conn.execute(run_configuration.table.insert(), rows)
                self._update_nested_tables_bulk(run_configuration, rows)
            self.lasteventids.update_eventid(self.conn, run_configuration.table_name, last_eventid)

    def load_events_from_file(self, events_path: str):
        """Load events from a file, primarily used for testing."""
        with open(events_path, "rb") as ef:
            for line in ef:
                if line := line.strip():
                    event_id, event_meta_str, data_str = line.split(b"|", maxsplit=2)
                    event_meta = orjson.loads(event_meta_str)
                    event_data = orjson.loads(data_str)
                    self.process_event(
                        event_meta,
                        event_data,
                    )

    def load_events_from_file_using_bulk(self, events_path: str):
        """Load events from a file, primarily used for testing."""
        with open(events_path, "rb") as ef:
            events = []
            for line in ef:
                if line := line.strip():
                    event_id, event_meta_str, data_str = line.split(b"|", maxsplit=2)
                    event_meta = orjson.loads(event_meta_str)
                    event_data = orjson.loads(data_str)
                    events.append((event_meta, event_data))
            self.process_events(events)



from schematools.factories import tables_factory


def test_through_col_creation(engine, brk_schema, verblijfsobjecten_schema):
    """Prove that through tables are containing all fields from the schema definition.

    When a schema has a relation that contains not only the (composite) foreign key fields
    but also extra fields that are defined on the relation
    (e.g. beginGeldigheid and eindGeldigheid.), those fields should also end up
    in the resulting SQLAlchemy tables.
    """
    sa_tables = tables_factory(brk_schema)

    for test_table_name in [
        "stukdelen_isBronVoorAantekeningKadastraalObject",
        "aantekeningenrechten_heeftBetrokkenPersoon",
    ]:

        colum_names = {c.name for c in sa_tables[test_table_name].columns}
        assert {"begin_geldigheid", "eind_geldigheid"} < colum_names



from schematools.introspect.geojson import geojson_to_table


def test_geojson_to_table():
    """Prove that GeoJSON can be parsed."""
    geojson = {
        "type": "FeatureCollection",
        "generator": "overpass-ide",
        "copyright": (
            "The data included in this document is from www.openstreetmap.org."
            " The data is made available under ODbL."
        ),
        "timestamp": "2019-11-15T10:27:02Z",
        "features": [
            {
                "type": "Feature",
                "properties": {
                    "@id": "relation/10191879",
                    "name": "gevaarlijke stoffen",
                    "route": "road",
                    "type": None,
                },
                "geometry": {
                    "type": "Point",
                    "coordinates": [[4.9377626, 52.3868045]],
                },
                "id": "relation/10191879",
            },
            {
                "type": "Feature",
                "properties": {
                    "@id": "relation/10191879",
                    "name": "gevaarlijke stoffen",
                    "route": "road",
                    "type": 3,
                },
                "geometry": {
                    "type": "Point",
                    "coordinates": [[4.9377626, 52.3868045]],
                },
                "id": "relation/10191879",
            },
        ],
    }

    result = geojson_to_table(iter(geojson["features"]), "unittest.geojson")
    assert result == [
        {
            "id": "unittest_relation",  # @id field was parsed as "relation"
            "type": "table",
            "version": "1.0.0",
            "schema": {
                "$schema": "http://json-schema.org/draft-07/schema#",
                "type": "object",
                "additionalProperties": False,
                "required": [],
                "display": "name",
                "properties": {
                    "schema": {
                        "$ref": (
                            "https://schemas.data.amsterdam.nl/schema@v1.3.0"
                            "#/definitions/schema"
                        )
                    },
                    "id": {"type": "string"},
                    "geometry": {"$ref": "https://geojson.org/schema/Point.json"},
                    "name": {"type": "string"},
                    "route": {"type": "string"},
                    "type": {"type": "number"},  # Introspected from 2nd record.
                },
            },
        }
    ]



from __future__ import annotations  # noqa: D104

from dataclasses import dataclass
from dataclasses import field as dataclass_field
from typing import Any, Callable, ClassVar
from urllib.parse import urlparse

from django.conf import settings
from factory import Faker as FactoryBoyFaker
from factory import LazyAttribute
from factory import Sequence as FactoryBoySequence
from factory.declarations import BaseDeclaration
from gisserver.geometries import CRS

from schematools.contrib.django.faker.providers import (  # noqa: F401, this is to load providers
    date,
    date_time,
    geo,
    integer,
    nuller,
    pyfloat,
)
from schematools.types import DatasetFieldSchema

LOCALE = "nl_NL"


@dataclass
class Declaration:
    """Sets up a BaseDeclaration and its parameters to be used for FactoryBoy.

    This wrapper class around BaseDeclaration enables a layered approach,
    where the parameters can be overridden at several levels.
    """

    cls: ClassVar[type] = BaseDeclaration
    arg: Any | None = None
    declaration_kwargs: dict[str, Any] = dataclass_field(default_factory=dict)
    field_kwargs: dict[str, Any] = dataclass_field(default_factory=dict)
    field_arg: Any | None = dataclass_field(init=False)

    def get_call_param_modifiers(self, field) -> tuple(Any, dict[str, Any]):
        """Get modifications to the args and kwarg for the __call__.

        When a Declaration is being used to create the actual mock data,
        for some mockers, additional or modified args or kwargs are needed.
        This method gives subclasses the opportunity to change arg or kwargs
        according to their needs.
        """
        return None, {}

    def __call__(self, field, **schema_faker_kwargs) -> BaseDeclaration:
        """Create the actual FactoryBoy declaration instance that is being used for mocking."""
        arg, kwargs = self.get_call_param_modifiers(field)
        call_kwargs = {}
        call_kwargs.update(self.declaration_kwargs)
        call_kwargs.update(kwargs)
        call_kwargs.update(schema_faker_kwargs)
        call_arg = arg or self.arg
        return self.cls(call_arg, **call_kwargs)


@dataclass
class Sequence(Declaration):
    """Declaration wrapper, wrapping the standard FactoryBoy Sequence."""

    cls: ClassVar = FactoryBoySequence
    arg: Callable[[], Any] = int


@dataclass
class Lazy(Declaration):
    """Declaration wrapper that is a base class for `LazyAttribute` based declarations."""

    cls: ClassVar = LazyAttribute


@dataclass
class PkJoiner(Lazy):
    """This declaration use the faker.LazyAttribute.

    So, the `self.arg` needs to be a function,
    that will be evaluated late in the mocking process.
    """

    def _join_field_values(self, obj):
        values = [str(getattr(obj, fid)) for fid in obj.table_schema.identifier]
        return ".".join(values)

    def __post_init__(self):
        self.arg = self._join_field_values


@dataclass
class Faker(Declaration):
    """Declarion wrapper base class for more specialized fakers."""

    cls: ClassVar = FactoryBoyFaker

    def __post_init__(self):
        self.declaration_kwargs["locale"] = LOCALE

    def get_call_param_modifiers(self, field) -> tuple(Any, dict[str, Any]):
        arg, kwargs = super().get_call_param_modifiers(field)
        if (elements := field.get("enum")) is not None:
            kwargs["elements"] = elements
            arg = "random_element"
        return arg, kwargs


@dataclass
class GeoFaker(Faker):
    """Declaration wrapper for the `geo` faker."""

    arg: str = "geo"  # The registered `geo` faker

    def get_call_param_modifiers(self, field) -> tuple(Any, dict[str, Any]):
        arg, kwargs = super().get_call_param_modifiers(field)
        return None, kwargs | {"crs": CRS.from_string(field.crs)}


@dataclass
class NullableFaker(Faker):
    def get_call_param_modifiers(self, field) -> tuple(Any, dict[str, Any]):
        arg, kwargs = super().get_call_param_modifiers(field)
        return None, kwargs | {"nullable": not field.required}


@dataclass
class NullableIntFaker(Faker):
    """Declaration wrapper for nullable integers.

    With support for `minimum` and `maximum`.
    """

    arg: str = "nullable_int"

    def get_call_param_modifiers(self, field) -> tuple(Any, dict[str, Any]):
        arg, kwargs = super().get_call_param_modifiers(field)
        if arg == "random_element":  # for enums, we return here
            return arg, kwargs
        if (min_ := field.get("minimum")) is not None:
            kwargs["min_"] = min_
        if (max_ := field.get("maximum")) is not None:
            kwargs["max_"] = max_
        return None, kwargs


@dataclass
class NullableFloatFaker(Faker):
    """Declaration wrapper for nullable floats.

    With support for `minimum` and `maximum`.
    """

    arg: str = "nullable_float"

    def get_call_param_modifiers(self, field) -> tuple(Any, dict[str, Any]):
        arg, kwargs = super().get_call_param_modifiers(field)
        if (min_value := field.get("minimum")) is not None:
            kwargs["min_value"] = min_value
        if (max_value := field.get("maximum")) is not None:
            kwargs["max_value"] = max_value
        return None, kwargs


DECLARATION_LOOKUP = {
    "string": Faker("pystr"),
    "integer": NullableIntFaker(),
    "integer/autoincrement": Sequence(),
    "string/autoincrement": Sequence(lambda n: str(n)),
    "number/autoincrement": Sequence(),
    "date": NullableFaker("nullable_date_object"),
    "date-time": NullableFaker("nullable_date_time"),
    "time": Faker("time"),
    "number": NullableFloatFaker(),
    "boolean": Faker("boolean"),
    "array": Faker("pylist", {"value_types": [str]}),
    "object": Faker("pystr"),  # needs a concatenated key field
    "json": Faker("json"),
    "/definitions/id": Faker("pyint"),
    "/definitions/schema": Faker("text"),
    "https://geojson.org/schema/Geometry.json": GeoFaker(
        declaration_kwargs={"geo_type": "Geometry"}
    ),
    "https://geojson.org/schema/Point.json": GeoFaker(declaration_kwargs={"geo_type": "Point"}),
    "https://geojson.org/schema/MultiPoint.json": GeoFaker(
        declaration_kwargs={"geo_type": "MultiPoint"}
    ),
    "https://geojson.org/schema/Polygon.json": GeoFaker(
        declaration_kwargs={"geo_type": "Polygon"}
    ),
    "https://geojson.org/schema/MultiPolygon.json": GeoFaker(
        declaration_kwargs={"geo_type": "MultiPolygon"}
    ),
    "https://geojson.org/schema/LineString.json": GeoFaker(
        declaration_kwargs={"geo_type": "LineString"}
    ),
    "https://geojson.org/schema/MultiLineString.json": GeoFaker(
        declaration_kwargs={"geo_type": "MultiLineString"}
    ),
    "https://geojson.org/schema/GeometryCollection.json": GeoFaker(
        declaration_kwargs={"geo_type": "GeometryCollection"},
    ),
    "city": Faker("city"),
    "sequence": Faker("sequence"),
    "bsn": Faker("ssn"),
    "street_name": Faker("street_name"),
    "nuller": Faker("nuller"),
    "joiner": PkJoiner(),
    "uri": Faker("uri"),
    "email": Faker("email"),
}


def get_field_factory(field: DatasetFieldSchema) -> BaseDeclaration:
    """Gets the appropriate field factory."""
    field_type = field.type
    # reduce amsterdam schema refs to their fragment
    if field_type.startswith(settings.SCHEMA_DEFS_URL):
        field_type = urlparse(field_type).fragment

    # If, in addition to a type, a format has been defined
    # for a field, the format will be used to look up the
    # appropriate provider.
    if (format_ := field.format) is not None:
        field_type = format_

    # If a faker has been defined for the field, this
    # faker is used for the provider lookup.
    schema_faker_kwargs = {}
    if (schema_faker := field.faker) is not None:
        field_type = schema_faker.name
        schema_faker_kwargs = schema_faker.properties
    declaration_maker = DECLARATION_LOOKUP.get(field_type)
    if declaration_maker is None:
        raise ValueError(f"No declaration defined for field with type {field_type}")

    return declaration_maker(field, **schema_faker_kwargs)



from __future__ import annotations

import os
from pathlib import Path

from psycopg2 import sql

from schematools.exports import _get_fields
from schematools.types import DatasetSchema


def export_geopackages(
    connection,
    dataset_schema: DatasetSchema,
    output: str,
    table_ids: list[str] | None = None,
    scopes: list[str] | None = None,
    size: int | None = None,
) -> None:
    """Export geopackages for all tables or an indicated subset in the dataset.

    Args:
        connection: SQLAlchemy connection object. Is needed for the `psycopg2`
        formatting.
        dataset_schema: Schema that needs export as geopackage.
        output: path on the filesystem where output should be stored.
        table_ids: optional parameter for a subset for the tables of the datasetself.
        scopes: Keycloak scopes that need to be taken into account.
            The geopackage will be produced contains information that is only
            accessible with these scopes.
        size: To produce a subset of the rows, mainly for testing.
    """

    base_dir = Path(output)
    db_url = connection.engine.url
    tables = (
        dataset_schema.tables
        if not table_ids
        else [dataset_schema.get_table_by_id(table_id) for table_id in table_ids]
    )
    command = 'ogr2ogr -f "GPKG" {output_path} PG:"{db_url}" -sql "{sql}"'
    for table in tables:
        output_path = base_dir / f"{table.db_name}.gpkg"
        field_names = sql.SQL(",").join(
            sql.Identifier(field.db_name)
            for field in _get_fields(dataset_schema, table, scopes)
            if field.db_name != "schema"
        )
        table_name = sql.Identifier(table.db_name)
        query = sql.SQL("SELECT {field_names} from {table_name}").format(
            field_names=field_names, table_name=table_name
        )
        if size is not None:
            query = f"{query} LIMIT {size}"
        sql_stmt = query.as_string(connection.connection.cursor())
        os.system(  # noqa: S605  # nosec: B605
            command.format(output_path=output_path, db_url=db_url, sql=sql_stmt)
        )



from __future__ import annotations

from pathlib import Path
from typing import IO

import psycopg2
from sqlalchemy import MetaData, Table
from sqlalchemy.engine import Connection

from schematools.factories import tables_factory
from schematools.types import _PUBLIC_SCOPE, DatasetFieldSchema, DatasetSchema, DatasetTableSchema

metadata = MetaData()


def enable_datetime_cast():
    """Register a special cast-to-string for the datetime type.

    This cast is not applied at module level, otherwise this
    cast would be enabled for all `schema-tools` code.
    """

    def cast_date(value, cur):
        if value is None:
            return None
        return str(value)

    # The bag_panden table had one record with a date (year = 0001)
    # that lets the dbapi `cursor.fetchmany` crash.
    # We need to define a custom type to handle this error,
    # because the historical record with this particular date cannot
    # be modified.
    date_oid = (1184,)  # This oid. can be obtained from cursor.description
    TIMESTAMPSTR = psycopg2.extensions.new_type(date_oid, "TIMESTAMP", cast_date)
    psycopg2.extensions.register_type(TIMESTAMPSTR)


class BaseExporter:
    """Baseclass for exporting tables rows."""

    extension = ""

    def __init__(
        self,
        connection: Connection,
        dataset_schema: DatasetSchema,
        output: str,
        table_ids: list[str] | None = None,
        scopes: list[str] | None = None,
        size: int | None = None,
    ):
        """Constructor.

        Args:
        connection: SQLAlchemy connection object.
        dataset_schema: Schema that needs export as geopackageself.
        output: path on the filesystem where output should be storedself.
        table_ids: optional parameter for a subset for the tables of the datasetself.
        scopes: Keycloak scopes that need to be taken into accountself.
            The geopackage will be produced contains information that is only
            accessible with these scopes.
        size: To produce a subset of the rows, mainly for testing.
        """
        self.connection = connection
        self.dataset_schema = dataset_schema
        self.table_ids = table_ids
        self.scopes = set(scopes)
        self.size = size

        self.base_dir = Path(output)
        self.tables = (
            dataset_schema.tables
            if not table_ids
            else [dataset_schema.get_table_by_id(table_id) for table_id in table_ids]
        )
        self.sa_tables = tables_factory(dataset_schema, metadata)

    def _get_column(self, sa_table: Table, field: DatasetFieldSchema):
        column = getattr(sa_table.c, field.db_name)
        processor = self.geo_modifier if field.is_geo else lambda col, _fn: col
        return processor(column, field.db_name)

    def _get_columns(self, sa_table: Table, table: DatasetTableSchema):
        for field in _get_fields(self.dataset_schema, table, self.scopes):
            try:
                yield self._get_column(sa_table, field)
            except AttributeError:
                pass  # skip unavailable columns

    def export_tables(self):
        for table in self.tables:
            srid = table.crs.split(":")[1] if table.crs else None
            if table.has_geometry_fields and srid is None:
                raise ValueError("Table has geo fields, but srid is None.")
            sa_table = self.sa_tables[table.id]
            with open(self.base_dir / f"{table.db_name}.{self.extension}", "w") as file_handle:
                self.write_rows(file_handle, table, sa_table, srid)

    def write_rows(
        self, file_handle: IO[str], table: DatasetTableSchema, sa_table: Table, srid: str
    ):
        raise NotImplementedError


def _get_fields(dataset_schema: DatasetSchema, table: DatasetTableSchema, scopes: list[str]):
    parent_scopes = set(dataset_schema.auth | table.auth) - {_PUBLIC_SCOPE}
    for field in table.fields:
        if field.is_array:
            continue
        if parent_scopes | set(field.auth) - {_PUBLIC_SCOPE} <= set(scopes):
            yield field



import re
from collections import defaultdict
from typing import Iterable, List, Optional

from django.core.management import BaseCommand, CommandError
from django.db import DatabaseError, connection, router, transaction

from schematools.contrib.django.factories import schema_models_factory
from schematools.contrib.django.models import Dataset


class Command(BaseCommand):
    help = "Create the tables based on the uploaded Amsterdam schema's."
    requires_system_checks = []  # don't test URLs (which create models)

    def add_arguments(self, parser):
        parser.add_argument(
            "--skip",
            dest="skip",
            nargs="*",
            help="Schemas that need to be skipped",
        )

    def handle(self, *args, **options):
        skip = options.get("skip")
        create_tables(self, Dataset.objects.db_enabled(), allow_unmanaged=True, skip=skip)


def create_tables(
    command: BaseCommand,
    datasets: Iterable[Dataset],
    allow_unmanaged: bool = False,
    base_app_name: Optional[str] = None,
    skip: Optional[List[str]] = None,
) -> None:  # noqa: C901
    """Create tables for all updated datasets.
    This is a separate function to allow easy reuse.
    """
    errors = 0
    command.stdout.write("Creating tables")

    # First create all models. This allows Django to resolve  model relations.
    models = []
    to_be_skipped = set(skip if skip is not None else [])

    # Because datasets are related, we need to 'prewarm'
    # the datasets cache (the DatasetSchema.dataset_collection)
    # by accessing the `Dataset.schema` attribute.
    for dataset in datasets:
        dataset.schema

    for dataset in datasets:
        if not dataset.enable_db or dataset.name in to_be_skipped:
            continue  # in case create_tables() is called by import_schemas

        models.extend(schema_models_factory(dataset, base_app_name=base_app_name))

    # Grouping multiple versions of same model by table name
    models_by_table = defaultdict(list)
    for model in models:
        models_by_table[model._meta.db_table].append(model)

    # Create all tables
    with connection.schema_editor() as schema_editor:
        for db_table_name, models_group in models_by_table.items():
            # Only create tables if migration is allowed
            # - router allows it (not some external database)
            # - table is not a view
            # - model is managed (not by default)
            # - user overrides this (e.g. developer)
            # - create table for latest version of this dataset group
            model = max(models_group, key=lambda model: model._dataset.version)

            is_view = False

            for view_table in model._dataset.schema.tables:
                if view_table.is_view:
                    command.stdout.write(f"  Skipping view: {db_table_name}")
                    is_view = True

            if is_view:
                continue

            router_allows = router.allow_migrate_model(model._meta.app_label, model)
            if not router_allows:
                command.stdout.write(f"  Skipping externally managed table: {db_table_name}")
                continue

            if not allow_unmanaged and not model._meta.can_migrate(connection):
                command.stderr.write(f"  Skipping non-managed model: {model._meta.db_table}")
                continue

            try:
                command.stdout.write(f"* Creating table {model._meta.db_table}")
                with transaction.atomic():
                    schema_editor.create_model(model)
            except (DatabaseError, ValueError) as e:
                command.stderr.write(f"  Tables not created: {e}")
                if not re.search(r'relation "[^"]+" already exists', str(e)):
                    errors += 1

    if errors:
        raise CommandError("Not all tables could be created")



from schematools.loaders import URLSchemaLoader


def test_load_all_publishers(schema_loader):
    pubs = schema_loader.get_all_publishers()
    assert pubs == {
        "GLEBZ": {
            "id": "GLEBZ",
            "name": "Datateam Glebz",
            "shortname": "braft",
            "tags": {"costcenter": "12345.6789"},
        },
        "HARRY": {
            "id": "HARRY",
            "name": "Datateam Harry",
            "shortname": "harhar",
            "tags": {"costcenter": "123456789.4321.13519", "team": "taggy"},
        },
        "NOTTHESAMEASFILENAME": {
            "id": "NOTTHESAMEASFILENAME",
            "name": "Datateam incorrect",
            "shortname": "nono",
            "tags": {"costcenter": "1236789.4321.13519", "team": "taggy"},
        },
    }


def test_publisher_url():
    loader = URLSchemaLoader("https://foo.bar/baz/datasets/")
    assert loader._get_publisher_url() == "https://foo.bar/baz/publishers"



from __future__ import annotations

from datetime import date, datetime

from factory import Faker
from faker.generator import Generator
from faker.providers import BaseProvider
from faker.providers.date_time import Provider as DateTimeProvider


class NullableDateProvider(BaseProvider):
    """Provider for a date or None provider.

    To be used for nullable fields in a database.

    """

    def __init__(self, generator: Generator) -> None:  # noqa: D107
        super().__init__(generator)
        self.date_time = DateTimeProvider(generator)

    def nullable_date_object(  # noqa: D102
        self,
        end_datetime: datetime = None,
        nullable=False,
    ) -> date | None:

        if nullable:
            if self.generator.random.randint(0, 1):
                return None
        return self.date_time.date_object(
            end_datetime=end_datetime,
        )


Faker.add_provider(NullableDateProvider, locale="nl_NL")



import json
import os
import pprint
import typing

import click
import jsonref


class JsonStrategicLoader(jsonref.JsonLoader):
    def __init__(self, strategies: typing.List[typing.Callable], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.strategies = strategies

    def _call_strategy(self, strategy: typing.Callable, uri: str, **load_params):
        return json.loads(strategy(uri), **load_params)

    def _call_parent(self, uri, **load_params):
        return super().__call__(uri, **load_params)

    def __call__(self, uri: str, **load_params):
        for strategy in self.strategies:
            try:
                return self._call_strategy(strategy, uri, **load_params)
            except ValueError:
                pass
        return self._call_parent(uri, **load_params)


def local_strategy(uri: str) -> str:
    if uri.startswith("./"):
        return open(uri).read()
    raise ValueError


def load(document_file: typing.IO[str], base_uri: str = "", loader=None, backend=jsonref.load):
    loader = loader or JsonStrategicLoader([local_strategy])
    return backend(document_file, base_uri, loader)


@click.command()
@click.argument("document_path", type=click.types.Path(), required=True)
def main(document_path):
    base_uri = f"file://{os.path.dirname(os.path.realpath(document_path))}/"
    with open(document_path) as document:
        result = load(document, base_uri)
        pprint.pprint(result)



import sqlite3

import orjson
import pytest

from schematools.exports.csv import export_csvs
from schematools.exports.geopackage import export_geopackages
from schematools.exports.jsonlines import export_jsonls
from schematools.importer.ndjson import NDJSONImporter


def _load_meetbouten_content(here, engine, meetbouten_schema):
    ndjson_path = here / "files" / "data" / "meetbouten.ndjson"
    importer = NDJSONImporter(meetbouten_schema, engine)
    importer.generate_db_objects("meetbouten", truncate=True, ind_extra_index=False)
    importer.load_file(ndjson_path)


def test_csv_export(here, engine, meetbouten_schema, dbsession, tmp_path):
    """Prove that csv export contains the correct content."""
    _load_meetbouten_content(here, engine, meetbouten_schema)
    export_csvs(engine, meetbouten_schema, str(tmp_path), [], [], 1)
    with open(tmp_path / "meetbouten_meetbouten.csv") as out_file:
        assert out_file.read() == (
            "identificatie,ligtInBuurtId,merkCode,merkOmschrijving,geometrie\n"
            "1,10180001.1,12,De meetbout,SRID=28992;POINT(119434 487091.6)\n"
        )


def test_jsonlines_export(here, engine, meetbouten_schema, dbsession, tmp_path):
    """Prove that jsonlines export contains the correct content."""
    _load_meetbouten_content(here, engine, meetbouten_schema)
    export_jsonls(engine, meetbouten_schema, str(tmp_path), [], [], 1)
    with open(tmp_path / "meetbouten_meetbouten.jsonl") as out_file:
        result = orjson.loads(out_file.read())
        result["geometrie"]["coordinates"][1] = round(result["geometrie"]["coordinates"][1], 5)
        result["geometrie"]["coordinates"][0] = round(result["geometrie"]["coordinates"][0], 5)
        assert result == {
            "identificatie": 1,
            "ligtInBuurtId": "10180001.1",
            "merkCode": "12",
            "merkOmschrijving": "De meetbout",
            "geometrie": {"type": "Point", "coordinates": [4.86497, 52.37055]},
        }


# We have to skip this test, ogr2og2 is not available on github
# We need to think of a flag to enable this test locally with `skipif`.
@pytest.mark.skip()
def test_geopackage_export(here, engine, meetbouten_schema, dbsession, tmp_path):
    """Prove that geopackage export contains the correct content."""
    _load_meetbouten_content(here, engine, meetbouten_schema)
    with engine.begin() as connection:
        export_geopackages(connection, meetbouten_schema, "/tmp", [], [])
        export_geopackages(connection, meetbouten_schema, str(tmp_path), [], [])
    sqlite3_conn = sqlite3.connect(tmp_path / "meetbouten_meetbouten.gpkg")
    cursor = sqlite3_conn.cursor()
    cursor.execute("select * from rtree_sql_statement_geometrie")
    res = cursor.fetchall()
    assert res == [(1, 119434.0, 119434.0, 487091.59375, 487091.65625)]
    cursor.execute(
        """
            select identificatie, ligt_in_buurt_id, merk_code, merk_omschrijving from sql_statement
       """
    )
    res = cursor.fetchall()
    assert res == [(1, "10180001.1", "12", "De meetbout")]



from typing import Iterable, Optional

from django.core.management import BaseCommand, CommandError
from django.db import DatabaseError, connection, transaction
from psycopg2 import sql

from schematools.contrib.django.models import Dataset, DatasetTableSchema
from schematools.naming import to_snake_case

DATASETS = Dataset.objects.db_enabled()


class Command(BaseCommand):
    help = "Create the views based on the uploaded Amsterdam schema's."
    requires_system_checks = []  # don't test URLs (which create models)

    def handle(self, *args, **options):
        create_views(self, Dataset.objects.db_enabled())


def _is_valid_sql(view_sql: str, view_name: str, write_role_name: str) -> bool:
    """Try and execute the SQL return true if it succeeds, rollback the transaction."""
    try:
        with transaction.atomic():
            sid = transaction.savepoint()
            with connection.cursor() as cursor:

                # Check if write role exists and create if it does not
                _create_role_if_not_exists(cursor, write_role_name)

                cursor.execute(
                    sql.SQL("SET ROLE {write_role_name}").format(
                        write_role_name=sql.Identifier(write_role_name)
                    )
                )

                cursor.execute(
                    sql.SQL("DROP VIEW IF EXISTS {view_name} CASCADE").format(
                        view_name=sql.Identifier(view_name)
                    )
                )
                cursor.execute(view_sql)
            transaction.savepoint_rollback(sid)
    except Exception as e:  # noqa: F841
        return False
    return True


def _get_scopes(datasetname: str, tablename: str) -> frozenset[str]:
    from functools import reduce
    from operator import __or__

    dataset = DATASETS.get(name=to_snake_case(datasetname)).schema
    if tablename in [table.id for table in dataset.tables]:
        table = dataset.get_table_by_id(tablename)
        return table.auth | reduce(__or__, [f.auth for f in table.fields])
    return frozenset()


def _get_required_permissions(
    table: DatasetTableSchema,
) -> frozenset[str]:
    derived_from = table.derived_from
    all_scopes = frozenset()
    for relation in derived_from:
        datasetname, tablename = relation.split(":")
        all_scopes |= _get_scopes(datasetname, tablename)
    return all_scopes


def _check_required_permissions_exist(
    view_dataset_auth: frozenset[str], required_permissions: frozenset[str]
) -> bool:
    """Check if the required permissions exist in the table and dataset auth.

    Args:
        view_dataset_auth (frozenset[str]): The auth parameter of the dataset the view is in.
        required_permissions (frozenset[str]): The required permissions for the view.

    returns:
        bool: True if all required permissions exist in the view dataset auth, False if not.

    """

    return required_permissions <= (view_dataset_auth | {"OPENBAAR"})


def _clean_sql(sql) -> str:
    """Clean the SQL to make it easier to parse."""

    sql = sql.replace("\n", " ").lower().strip()

    if sql[-1] != ";":
        sql += ";"

    return sql


def _create_role_if_not_exists(cursor, role_name):
    # Create the role if it doesn't exist
    cursor.execute(
        sql.SQL("SELECT 1 FROM pg_roles WHERE rolname={role_name}").format(
            role_name=sql.Literal(role_name)
        )
    )

    role_exists = cursor.fetchone()
    if not role_exists:
        cursor.execute(
            sql.SQL("CREATE ROLE {role_name}").format(role_name=sql.Identifier(role_name))
        )


def create_views(
    command: BaseCommand,
    datasets: Iterable[Dataset],
    base_app_name: Optional[str] = None,
) -> None:
    """Create views. This is a separate function to allow easy reuse."""
    errors = 0
    command.stdout.write("Creating views")

    # Because datasets are related, we need to 'prewarm'
    # the datasets cache (the DatasetSchema.dataset_collection)
    # by accessing the `Dataset.schema` attribute.
    for dataset in datasets:
        dataset.schema

    for dataset in datasets:
        if not dataset.enable_db:
            continue  # in case create_views() is called by import_schemas

    # Create all views
    for dataset in datasets:
        for table in dataset.schema.tables:
            if table.is_view:
                command.stdout.write(f"* Creating view {table.db_name}")

                # Generate the write role name
                write_role_name = f"write_{table._parent_schema.db_name}"

                # Check if the view sql is valid
                # If not skip this view and proceed with next view
                view_sql = _clean_sql(dataset.schema.get_view_sql())
                if not _is_valid_sql(view_sql, table.db_name, write_role_name):
                    command.stderr.write(f"  Invalid SQL for view {table.db_name}")
                    continue

                required_permissions = _get_required_permissions(table)
                view_dataset_auth = dataset.schema.auth
                if _check_required_permissions_exist(view_dataset_auth, required_permissions):
                    try:
                        with connection.cursor() as cursor:

                            # Check if write role exists and create if it does not
                            _create_role_if_not_exists(cursor, write_role_name)

                            # Grant usage and create on schema public to write role
                            cursor.execute(
                                sql.SQL(
                                    "GRANT usage,create on schema public TO {write_role_name}"
                                ).format(write_role_name=sql.Identifier(write_role_name))
                            )

                            # We create one `view_owner` role that owns all views
                            _create_role_if_not_exists(cursor, "view_owner")

                            cursor.execute("GRANT view_owner TO current_user")

                            cursor.execute(
                                sql.SQL("GRANT {write_role_name} TO view_owner").format(
                                    write_role_name=sql.Identifier(write_role_name)
                                )
                            )

                            # Loop though all required permissions and and grant them to the
                            # write user
                            for scope in required_permissions:
                                scope = f'scope_{scope.replace("/", "_").lower()}'
                                if scope:
                                    cursor.execute(
                                        sql.SQL("GRANT {scope} TO {write_role_name}").format(
                                            scope=sql.Identifier(scope),
                                            write_role_name=sql.Identifier(write_role_name),
                                        )
                                    )

                            cursor.execute(
                                sql.SQL("GRANT scope_openbaar TO {write_role_name}").format(
                                    write_role_name=sql.Identifier(write_role_name)
                                )
                            )

                            # Set the role before creating the view because the view is created
                            # with the permissions of the role
                            cursor.execute(
                                sql.SQL("SET ROLE {write_role_name}").format(
                                    write_role_name=sql.Identifier(write_role_name)
                                )
                            )

                            # Remove the view if it exists
                            cursor.execute(
                                sql.SQL("DROP VIEW IF EXISTS {view_name} CASCADE").format(
                                    view_name=sql.Identifier(table.db_name)
                                )
                            )

                            # Create the view
                            cursor.execute(view_sql)

                            # Reset the role to the default role
                            cursor.execute("RESET ROLE")

                            # Remove create and usage from write role
                            cursor.execute(
                                sql.SQL(
                                    "REVOKE usage,create on schema public FROM {write_role_name}"
                                ).format(write_role_name=sql.Identifier(write_role_name))
                            )

                            cursor.close()
                    except (DatabaseError, ValueError) as e:
                        command.stderr.write(f"  View not created: {e}")
                        errors += 1
                else:
                    command.stderr.write(
                        f"  Required permissions {required_permissions} not found in view auth"
                    )

    if errors:
        raise CommandError("Not all views could be created")



from typing import Collection, Optional

from django.db import connection

from .factories import schema_models_factory
from .models import Dataset


def create_tables(
    dataset: Dataset,
    tables: Optional[Collection[str]] = None,
    base_app_name: Optional[str] = None,
):
    """Create the database tables for a given schema."""
    with connection.schema_editor() as schema_editor:
        for model in schema_models_factory(dataset, tables=tables, base_app_name=base_app_name):
            schema_editor.create_model(model)



from __future__ import annotations

import csv
from typing import IO

from geoalchemy2 import functions as func  # ST_AsEWKT
from sqlalchemy import MetaData, Table, select
from sqlalchemy.engine import Connection

from schematools.exports import BaseExporter, enable_datetime_cast
from schematools.naming import toCamelCase
from schematools.types import DatasetSchema

metadata = MetaData()


class CsvExporter(BaseExporter):  # noqa: D101
    extension = "csv"
    geo_modifier = staticmethod(lambda col, fn: func.ST_AsEWKT(col).label(fn))

    def write_rows(  # noqa: D102
        self, file_handle: IO[str], table: DatasetSchema, sa_table: Table, srid: str
    ):
        columns = list(self._get_columns(sa_table, table))
        field_names = [c.name for c in columns]
        writer = csv.DictWriter(file_handle, field_names, extrasaction="ignore")
        writer.writerow({fn: toCamelCase(fn) for fn in field_names})
        query = select(self._get_columns(sa_table, table))
        if self.size is not None:
            query = query.limit(self.size)
        result = self.connection.execution_options(yield_per=1000).execute(query)
        for partition in result.partitions():
            for r in partition:
                writer.writerow(dict(r))


def export_csvs(
    connection: Connection,
    dataset_schema: DatasetSchema,
    output: str,
    table_ids: list[str],
    scopes: list[str],
    size: int,
):
    """Utility function to wrap the Exporter."""
    enable_datetime_cast()
    exporter = CsvExporter(connection, dataset_schema, output, table_ids, scopes, size)
    exporter.export_tables()
